// PROFILE.C
//
// Profile management for X36 driver.
// 


// #define DBG 1

#include <WDM.H>
#include <filter.h>





PSCRIPT ParseScript(IN PVOID Data, IN ULONG Length) // Converts a reg key into a script.
{
	PSCRIPT Head = NULL;
	PSCRIPT Current,Previous = NULL;
	PVOID Pos;
	unsigned long i=0;
	unsigned long j=0;
	
//	DbgPrint("ParseScript:  Length: %d\n",Length);	
	Pos=Data;
	Head = (PSCRIPT) ExAllocatePool(PagedPool, sizeof(SCRIPT));

	Current = Head;
	for (i=0; i+3<Length; i+=4)
	{
//		DbgPrint("ParseScript: count %d of length %d\n",i,Length);
		if (Current == NULL)
			i=Length;
		else
		{	
			RtlZeroMemory(Current,sizeof(SCRIPT));
			RtlCopyMemory(Current,Pos,2);
			Pos=(unsigned char*)Pos+2;
			//?
			j = (unsigned long) *(unsigned short*)((unsigned char*)Pos);
			Current->Delay = j * 10000;
			Pos=(unsigned char*)Pos+2;
//			DbgPrint("ParseScript:  Adding delay: %d   Offset: %d    Op: %x\n",Current->Delay, Current->Offset, Current->Op);	
			Current->Next=ExAllocatePool(PagedPool, sizeof(SCRIPT));
			Previous=Current;
			Current=Current->Next;
		}
	}
	if(Current != NULL)
	{
		ExFreePool(Current);
		if (Previous)
			Previous->Next = NULL;
		else
			Head=NULL;
	}
	
	return Head;
}	
	


 NTSTATUS AddEntry ( IN PWSTR ValueName, IN ULONG ValueType, IN PVOID ValueData, IN ULONG ValueLength, IN PVOID Context, IN PVOID EntryContext )
{
	unsigned short devicetype, number, command, switch1, switch2, shift, direction;
	struct DEVICE_EXTENSION *devExt = (struct DEVICE_EXTENSION *) Context;
	UNICODE_STRING  Name;
	
//	DbgPrint("AddEntry:  Name: %S   Data: %X\n",ValueName,*(unsigned int *)ValueData);
	
	if ((ValueType != REG_BINARY) || (ValueName[0]==0))
		return STATUS_SUCCESS;
	
	RtlInitUnicodeString(&Name,ValueName);
	if (RtlUnicodeStringToAnsiSize(&Name)<7)
	{
		return STATUS_SUCCESS;
	}
	
	
	// First, parse key name.
	devicetype = ValueName[0];
	number = ValueName[1]-48;
	command = ValueName[2];
	switch1 = ValueName[3]-48;
	switch2 = ValueName[4]-48;
	shift = ValueName[5]-48;
	
	// Sanity check.
//	DbgPrint("Sw1: %d    Sw2: %d    Shift: %d\n",switch1,switch2,shift);
	if ((switch1 > 3) || (switch1 < 0 ) || (switch2>3) || (switch2<0) || (shift > 1) || (shift < 0))
	{
		return STATUS_SUCCESS;		
	}
	

	if (devExt->Configurations[switch1][switch2][shift] == NULL) 
	{
		devExt->Configurations[switch1][switch2][shift] = ExAllocatePool(PagedPool, sizeof(CONFIG));
		if (devExt->Configurations[switch1][switch2][shift] == NULL) 
			return STATUS_INSUFFICIENT_RESOURCES;
		RtlZeroMemory(devExt->Configurations[switch1][switch2][shift], sizeof(CONFIG));
	}
	
	switch (devicetype)
	{
		case 'a': // Axis.
			// Sanity check.
			if ((number > 5) || (number < 0) || (ValueLength != 2)) return STATUS_SUCCESS;
		
			switch (command)
			{
				case 'x': // Maxiumum.
					devExt->Configurations[0][0][0]->Axis[number].Max=*(unsigned short*)ValueData;
					break;
					
				case 'n': // Minimum.
					devExt->Configurations[0][0][0]->Axis[number].Min=*(unsigned short*)ValueData;
					break;
					
				case 'm': // Multiplier.
					break;
					
				case 'u': // Up.
					break;
					
				case 'd': // Down.
					break;
					
				default: // Nothing.
					;
			}
			break;
		
		case 'b': // Button.
			// Sanity check.
			if ((number > 7) || (number < 0)) return STATUS_SUCCESS;

			switch (command)
			{
				case 'p': // Pressed.
					devExt->Configurations[switch1][switch2][shift]->Button[number].Pressed = ParseScript(ValueData, ValueLength);
					break;
					
				case 'r': // Released.
					devExt->Configurations[switch1][switch2][shift]->Button[number].Released = ParseScript(ValueData, ValueLength);
					break;
					
				case 'h': // Hold.
					devExt->Configurations[switch1][switch2][shift]->Button[number].Hold = ParseScript(ValueData, ValueLength);
					break;
					
				default: // Nothing.					
					;
			}		
			break;
			
		case 's': // Switch.
			// Sanity check.
			if ((number > 1) || (number < 0)) return STATUS_SUCCESS;
		
		
			switch (command)
			{
				case 'p': // Pressed.
					devExt->Configurations[switch1][switch2][shift]->Slider[number].Pressed = ParseScript(ValueData, ValueLength);				
					break;
					
				case 'r': // Released.
					devExt->Configurations[switch1][switch2][shift]->Slider[number].Released = ParseScript(ValueData, ValueLength);				
					break;
					
				case 'h': // Hold.
					devExt->Configurations[switch1][switch2][shift]->Slider[number].Hold = ParseScript(ValueData, ValueLength);				
					break;
					
				default: // Nothing.					
					;
			}				
			break;
		
		case 'h': // Hat.
			// Sanity check.
			if ((number > 3) || (number < 0)) return STATUS_SUCCESS;

			// Get direciton.
			direction=ValueName[6]-48;
			if ((direction > 8) || (direction < 0)) return STATUS_SUCCESS;
			
			switch (command)
			{
				case 'p': // Pressed.
					devExt->Configurations[switch1][switch2][shift]->Hat[number].Pressed[direction] = ParseScript(ValueData, ValueLength);				
					break;
					
				case 'r': // Released.
					devExt->Configurations[switch1][switch2][shift]->Hat[number].Released[direction] = ParseScript(ValueData, ValueLength);				
					break;
					
				case 'h': // Hold.
					devExt->Configurations[switch1][switch2][shift]->Hat[number].Hold[direction] = ParseScript(ValueData, ValueLength);				
					break;

				case 'm': // Mode
					devExt->Configurations[switch1][switch2][shift]->Hat[number].Mode = *(unsigned char*)ValueData; // Parse data somehow?
					
				default: // Nothing.					
					;
			}				
			break;
			
		default:
		
			// Do nothing.
			;
	}
//	DbgPrint("AddEntry complete.\n");
	return STATUS_SUCCESS;	
}

 

NTSTATUS LoadProfile(struct DEVICE_EXTENSION  *DeviceExtension)
{
	RTL_QUERY_REGISTRY_TABLE myKey[2];
	NTSTATUS status;
	
//	DbgPrint("LoadProfile Begins.\n");
	
	myKey[0].QueryRoutine=AddEntry;
	myKey[0].Flags=RTL_QUERY_REGISTRY_REQUIRED;
	myKey[0].Name=NULL;
	myKey[0].EntryContext=NULL;
	myKey[0].DefaultType=0;
	myKey[0].DefaultData=NULL;
	myKey[0].DefaultLength=0;
	myKey[1].QueryRoutine=NULL;
	myKey[1].Name=NULL;
	
	
	status=RtlQueryRegistryValues( RTL_REGISTRY_CONTROL, L"X36", myKey, (PVOID) DeviceExtension, NULL );
	
//	DbgPrint("LoadProfile ends.  Status: %x\n",status);
	return status;
}


void DestroyScript(PSCRIPT Script)
{
	PSCRIPT Current=Script;
	PSCRIPT Next;
	
	while (Current != NULL)
	{
		Next=Current->Next;
		ExFreePool(Current);
		Current=Next;
	}
}
	

NTSTATUS ClearProfile(struct DEVICE_EXTENSION  *DeviceExtension) // Clears a profile and destroys all memory structures.
{
	unsigned int i,j,k,l,m;
	
	for(i=0;i<4;i++)
		for(j=0;j<4;j++)
			for(k=0;k<3;k++)
			{
//				DbgPrint("Now clearing profile %d %d %d\n",i,j,k);
				if (DeviceExtension->Configurations[i][j][k] != NULL)
				{
					// Axes
					;					
					// Buttons
					for(l=0;l<8;l++)
					{
//						DbgPrint("Button %d pressed\n",l);
						DestroyScript(DeviceExtension->Configurations[i][j][k]->Button[l].Pressed);
//						DbgPrint("Button %d released\n",l);
						DestroyScript(DeviceExtension->Configurations[i][j][k]->Button[l].Released);
//						DbgPrint("Button %d hold\n",l);
						DestroyScript(DeviceExtension->Configurations[i][j][k]->Button[l].Hold);
					}
					// Switches
					for(l=0;l<2;l++)
					{
//						DbgPrint("Switch %d pressed\n",l);
						DestroyScript(DeviceExtension->Configurations[i][j][k]->Slider[l].Pressed);
//						DbgPrint("Switch %d released\n",l);
						DestroyScript(DeviceExtension->Configurations[i][j][k]->Slider[l].Released);
//						DbgPrint("Switch %d hold\n",l);
						DestroyScript(DeviceExtension->Configurations[i][j][k]->Slider[l].Hold);
					}					
					// Hat
					for(l=0;l<4;l++)
					{
						for(m=0;m<9;m++)
						{
//							DbgPrint("Hat %d Dir %d pressed\n",l,m);
							DestroyScript(DeviceExtension->Configurations[i][j][k]->Hat[l].Pressed[m]);
//							DbgPrint("Hat %d Dir %d released\n",l,m);
							DestroyScript(DeviceExtension->Configurations[i][j][k]->Hat[l].Released[m]);
//							DbgPrint("Hat %d Dir %d hold\n",l,m);
							DestroyScript(DeviceExtension->Configurations[i][j][k]->Hat[l].Hold[m]);
						}
					}					
//					DbgPrint("Freeing memory\n");
					ExFreePool(DeviceExtension->Configurations[i][j][k]);
					DeviceExtension->Configurations[i][j][k]=NULL;
				}
			};;;
			
	// Init config 0.
//	DbgPrint("Allocating config 0\n");
	DeviceExtension->Configurations[0][0][0] = ExAllocatePool(PagedPool, sizeof(CONFIG));
	if (DeviceExtension->Configurations[0][0][0] == NULL) 
			return STATUS_INSUFFICIENT_RESOURCES;
//	DbgPrint("Clearing config 0\n");			
	RtlZeroMemory(DeviceExtension->Configurations[0][0][0], sizeof(CONFIG));		
	return STATUS_SUCCESS;
}


NTSTATUS ReloadProfile(struct DEVICE_EXTENSION  *DeviceExtension) // Checks to see if a new profile is necessary and reloads if it is.
{
	HANDLE myKey;
	NTSTATUS status;
	ULONG value=0;
	UNICODE_STRING ValueName;
	
    status = IoOpenDeviceRegistryKey(   DeviceExtension->physicalDevObj, 
                                        PLUGPLAY_REGKEY_DEVICE, 
                                        KEY_READ, 
                                        &myKey);

	if (! NT_SUCCESS(status))
		return status;
		
	
	status = QueryDeviceKey( myKey, L"LoadProfile",&value,sizeof(value));
	
	if ((NT_SUCCESS(status))&&(value))
	{		
//		DbgPrint("Reload required.  Reloading profile.\n");
		// Set reg value to 0.
		value=0;
		
		RtlInitUnicodeString(&ValueName,L"LoadProfile");
		
		ZwSetValueKey( myKey, &ValueName, 0, REG_BINARY, &value, sizeof(&value));
		
		// Clear current profile.
		ClearProfile(DeviceExtension);
		// Load new profile.
		LoadProfile(DeviceExtension);
		
//		DbgPrint("Reload complete.  Reloading profile.\n");
	}
	ZwClose(myKey);

//	DbgPrint("ReloadProfile ends.  Status: %x\n",status);
	return status;
}
